'// UNAME - Makes all files in a subdirectory have unique names
'// Created 2005 by Casey Sprangel
'// Copyright 2005, Casey Sprangel, all rights reserved.
'// Written for PowerBasic CC 4.0 (www.powerbasic.com)
'//---------------------------------------------------------------------------

#COMPILE EXE                                                    '// make a program
#DIM ALL                                                        '// force all functions and variables to be declared before use
#INCLUDE "WIN32API.INC"                                         '// include Windows API calls for the directory scanning functions
DECLARE FUNCTION exist(filen AS STRING) AS LONG                 '// declare the local functions we use
DECLARE FUNCTION isdir(filen AS STRING) AS LONG
$title = "UNAME "                                               '// static string for title bar display

'//---------------------------------------------------------------------------

FUNCTION isdir(filen AS STRING) AS LONG
  '// Returns -1 if a file or path exists and is a directory. Returns 0 otherwise.
  '// Does not change DIR$ FindFirst/FindNext pointers.
  DIM result AS LOCAL LONG                                      '// a junk variable, we don't care what comes back

  IF RTRIM$(filen, "\") = "." THEN                              '// the place holder for the current directory?
    FUNCTION = -1                                               '// it exists
    EXIT FUNCTION                                               '// leave function, we are done
  END IF
  IF RIGHT$(filen, 1) = ":" THEN                                '// does file name end in a colon (i.e. a drive letter)?
    result = GETATTR(filen + "\")                               '// see if the root path exists (adding a slash)
  ELSE                                                          '// this is a normal path or filename
    result = GETATTR(RTRIM$(filen, "\"))                        '// see if it exists, paths with trailing slash marks are trimmed
  END IF
  IF (ERR = 0) THEN                                             '// does it exist?
    IF (result AND 16&) = 16& THEN                              '// is this a directory?
      FUNCTION = -1                                             '// return true, this is a directory
    ELSE
      FUNCTION = 0                                              '// return false, this is not a directory
    END IF
  ELSE                                                          '// file does not exist
    ERRCLEAR                                                    '// clear the error flag
    FUNCTION = 0                                                '// return false
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION exist(filen AS STRING) AS LONG
  '// Returns -1 if a file or path exists. Returns 0 if it does not exist.
  '// Does not change DIR$ FindFirst/FindNext pointers
  LOCAL result AS LONG                                          '// a junk variable, we don't care what comes back

  IF RIGHT$(filen, 1) = ":" THEN                                '// does file name end in a colon (i.e. a drive letter)?
    result = GETATTR(filen + "\")                               '// see if the root path exists (adding a slash)
  ELSE                                                          '// this is a normal path or filename
    result = GETATTR(RTRIM$(filen, "\"))                        '// see if it exists, paths with trailing slash marks are trimmed
  END IF
  FUNCTION = (ERRCLEAR = 0)                                     '// check to see if we had an error
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION PBMAIN () AS LONG                                      '// program starts here
  DIM b1 AS STRING                                              '// working string buffer for file name compare
  DIM b2 AS STRING                                              '// working string buffer for file name compare
  DIM buf AS STRING                                             '// generic working string buffer
  DIM change_count AS LONG                                      '// counts the number of changes in each loop through the file names
  DIM change_total AS LONG                                      '// counts the total number of file names changed
  DIM dbuf(10) AS STRING                                        '// start with room for 10 directories
  DIM dbuf_count AS LONG                                        '// directory total count
  DIM dbuf_point AS LONG                                        '// directory pointer
  DIM fbuf(1000) AS STRING                                      '// start with room for 1000 file names
  DIM fbuf_count AS LONG                                        '// find name total count
  DIM hd1 AS LONG                                               '// handle to directory we are scanning
  DIM i AS LONG                                                 '// generic counter
  DIM lpFindFileData AS WIN32_FIND_DATA                         '// structure for getting file info from Windows API
  DIM mytick AS STRING                                          '// holds the clock value to limit display updates (for higher performance)
  DIM old_dbuf_count AS LONG                                    '// remembers previous directory count when scanning for new directories
  DIM old_fbuf_count AS LONG                                    '// remembers previous file name count when scanning new directories
  DIM path AS STRING                                            '// holds the path to the file or directory while processing
  DIM tcount AS LONG                                            '// tracks the total run time for this program
  DIM froot AS STRING                                           '// holds the root file name
  DIM fext AS STRING                                            '// holds the file extension for rename counting
  DIM fcount AS LONG                                            '// counter to hold unique file number
  DIM notnum AS STRING                                          '// holds everything but allowed numbers

  STDOUT "Casey's UNAME version 20051118"                       '// show program and version info
  buf = COMMAND$                                                '// get the command line from the user
  IF INSTR(buf, "?") > 0 THEN                                   '// did they ask for help?
    STDOUT "This tool will scan a directory and all subdirectories to make sure"
    STDOUT "all of the files in the directory tree have unique names."
    STDOUT ""
    STDOUT "WARNING: USE ONLY WITH EXTREME CAUTION!"
    STDOUT ""
    STDOUT "Syntax: UNAME <path>"
    STDOUT "The special case path of a single period (.) can be used to point"
    STDOUT "to the current default directory if needed."
    EXIT FUNCTION                                               '// leave the program, we are done
  END IF
  REPLACE "/" WITH "\" IN buf                                   '// allow unix style paths to work (something nice for my friends)
  IF buf = "." THEN                                             '// short hand for this directory
    buf = CURDIR$                                               '// use the current directory
  END IF
  IF isdir(buf) THEN                                            '// is this a valid directory?
    path = buf                                                  '// remember starting path
  ELSE                                                          '// not a valid directory
    IF buf <> "" THEN                                           '// did user give us a path?
      STDOUT "Error: path (" + buf + ") not found."             '// tell them we can't find it
    ELSE                                                        '// no directory provided
      STDOUT "Error: you must provide a starting directory. Use /? for help."
    END IF
    EXIT FUNCTION
  END IF
  tcount = TIMER                                                '// remember the time this program started
  dbuf(0) = RTRIM$(path, "\")                                   '// remember starting path
  dbuf_point = 0                                                '// start by pointing to this directory
  STDOUT "Scanning " + dbuf(0)                                  '// tell user what we are doing
  GOSUB show_scan
  DO                                                            '// start processing loop
    path = dbuf(dbuf_point) + "\"                               '// get path with trailing backslash mark
    hd1 = FindFirstFile(path + "*.*", lpFindFileData)           '// find first file in this directory
    IF hd1 <> %INVALID_HANDLE_VALUE THEN                        '// did we get a valid directory handle?
      DO                                                        '// start working our loop
        IF (lpFindFileData.dwFileAttributes AND %FILE_ATTRIBUTE_DIRECTORY) = %FILE_ATTRIBUTE_DIRECTORY THEN '// is this a directory?
          IF lpFindFileData.cFileName = "." OR lpFindFileData.cFileName = ".." THEN '// did we find the placeholder entries?
            IF FindNextFile(hd1, lpFindFileData) = 0 THEN       '// did we fail to find another file in this directory?
              EXIT LOOP                                         '// leave, no more files to find in this directory
            END IF
            ITERATE                                             '// go work on next file
          END IF
          buf = path + lpFindFileData.cFileName                 '// move the complete file name into our string buffer
          INCR dbuf_count                                       '// add one to our directory counter
          IF dbuf_count > UBOUND(dbuf()) THEN                   '// no room for this directory?
            REDIM PRESERVE dbuf(dbuf_count + 100)               '// make room to grow
          END IF
          dbuf(dbuf_count) = buf                                '// add this directory to our list of known directories
        ELSE                                                    '// this is not a directory, so it must be a file
          IF LEN(lpFindFileData.cFileName) > 0 THEN             '// is the name NOT empty? (due to disk corruption)
            buf = lpFindFileData.cFileName                      '// get the file name
            IF INSTR(buf, "\") > 0 THEN                         '// is there a path part of this file name?
              buf = RIGHT$(buf, INSTR(-1, buf, "\") + 1)        '// keep just the file name part
            END IF
            INCR fbuf_count                                     '// add one to our file name counter
            GOSUB show_scan
            IF fbuf_count > UBOUND(fbuf()) THEN                 '// no room to add this file name?
              REDIM PRESERVE fbuf(fbuf_count + 1000)            '// make plenty of room to grow
            END IF
            fbuf(fbuf_count) = buf + $NUL + FORMAT$(dbuf_point) '// add file entry pointing to this dir
          END IF                                                '// name not empty?
        END IF                                                  '// is this a directory
        IF FindNextFile(hd1, lpFindFileData) = 0 THEN           '// no more files in this directory?
          EXIT LOOP                                             '// leave the loop
        END IF
      LOOP
      FindClose hd1                                             '// we are done with this directory, close the handle
      STDOUT "+ " + dbuf(dbuf_point) + $SPC + FORMAT$(fbuf_count - old_fbuf_count) + " file" + IIF$(fbuf_count - old_fbuf_count = 1, "","s") + _
             IIF$(dbuf_count - old_dbuf_count > 0, ", " + FORMAT$(dbuf_count - old_dbuf_count) + " director" + IIF$(dbuf_count - old_dbuf_count = 1, "y","ies"),"") + _
             " found."
      old_fbuf_count = fbuf_count
      old_dbuf_count = dbuf_count
      INCR dbuf_point                                           '// move to next known directory
    END IF
  LOOP WHILE dbuf_point <= dbuf_count                           '// loop until we scann all directories

  '// by the time we get here, we should know about all directories and all files
  REDIM PRESERVE dbuf(dbuf_count)                               '// trim tables to just the size we need
  REDIM PRESERVE fbuf(fbuf_count)
  STDOUT "Searching for duplicate file names."
  notnum = CHR$(0 TO 47, 58 TO 94, 96 TO 255)
  DO
    GOSUB show_fix
    ARRAY SORT fbuf(), COLLATE UCASE                            '// sort files by name (case insensitive)
    change_total = change_total + change_count                  '// remember all of the changes we make
    change_count = 0                                            '// start counter over at zero for this loop
    i = 1                                                       '// start with the second file name (0 is first)
    DO WHILE i <= fbuf_count                                    '// loop through all file names
      b1 = LEFT$(fbuf(i - 1),INSTR(fbuf(i - 1) + $NUL, $NUL) -1) '// get previous file name entry
      b2 = LEFT$(fbuf(i),INSTR(fbuf(i) + $NUL, $NUL) -1)        '// get current file name entry
      IF UCASE$(b1) = UCASE$(b2) THEN                           '// is this a duplicate file name? (case insensitive)
        froot = b1                                              '// make a copy of the file name
        IF INSTR(froot, ".") > 0 THEN                           '// is there a period in this file name?
          fext = MID$(froot, INSTR(-1, froot, "."))             '// remember extension including period
          froot = LEFT$(froot, INSTR(-1, froot, ".") -1)        '// drop everything after last period
        ELSE
          fext = ""                                             '// there is no extension for this file
        END IF
        IF RIGHT$(froot, 1) = "]" THEN                          '// does this filename end with a counter in square brackets?
          IF INSTR(froot, "[") > 0 THEN                         '// is there an open counter bracket?
            IF INSTR(MID$(froot, INSTR(-1, froot, "[") + 1, LEN(froot) - INSTR(-1, froot, "[") -1), ANY notnum) = 0 THEN '// only numbers?
              froot = LEFT$(froot, INSTR(-1, froot, "[") - 1)   '// cut off the counter so we can add our own
            END IF
          END IF
        END IF
'        IF RIGHT$(froot, 1) = ")" THEN                          '// does this filename end with a counter in parens?
'          IF INSTR(froot, "(") > 0 THEN                         '// is there an open counter paren?
'            IF INSTR(MID$(froot, INSTR(-1, froot, "(") + 1, LEN(froot) - INSTR(-1, froot, "(") -1), ANY notnum) = 0 THEN '// only numbers?
'              froot = LEFT$(froot, INSTR(-1, froot, "(") - 1)   '// cut off the counter so we can add our own
'            END IF
'          END IF
'        END IF
        path = dbuf(VAL( MID$(fbuf(i - 1),INSTR(fbuf(i - 1) + $NUL, $NUL) + 1))) + "\" '// get directory
        DO WHILE exist(path + froot + "[" + FORMAT$(fcount) + "]" + fext) '// loop until we find a unique name
          INCR fcount                                           '// add one to our counter
        LOOP
        buf = froot + "[" + FORMAT$(dbuf_point) + "]" + fext    '// build new filename
        IF UCASE$(b1) <> UCASE$(buf) THEN                       '// did we find a new name works better?
          STDOUT path + b1 + " -> " + buf                       '// show old and new file name to user
          NAME path + b1 AS path + buf                          '// rename the file
          buf = buf + $NUL + MID$(fbuf(i - 1), INSTR(fbuf(i - 1) + $NUL, $NUL) + 1) '// build changed file record
          fbuf(i - 1) = buf                                     '// remember change in ram table for later compares
          INCR change_count                                     '// remember that we changed a file name in this loop
          GOSUB show_fix                                        '// update the screen
        END IF
      END IF
      INCR i                                                    '// move pointer to next file name
    LOOP
  LOOP UNTIL change_count = 0                                   '// no changes made in this loop through the file names?
  STDOUT "Found " + FORMAT$(fbuf_count) + " file" + IIF$(fbuf_count = 1, "", "s") + _
         " in " + FORMAT$(dbuf_count + 1) + " director" + IIF$(dbuf_count = 0, "y", "ies") + "."
  STDOUT "Changed " + FORMAT$(change_total) + " file name" + IIF$(change_total = 1,"","s")
  STDOUT "Runtime " + FORMAT$(TIMER - tcount,"0.") + " seconds."
  EXIT FUNCTION

show_scan:
  IF mytick <> RIGHT$(TIME$, 1) THEN                            '// did the clock change?
    CONSOLE NAME $title + dbuf(dbuf_point) + " (" + FORMAT$(fbuf_count) + " file" + IIF$(fbuf_count = 1,"","s") + ")"
    mytick = RIGHT$(TIME$, 1)                                   '// remember this second
  END IF
  RETURN
show_fix:
  IF mytick <> RIGHT$(TIME$, 1) THEN                            '// did the clock change?
    CONSOLE NAME $title + "Changed " + FORMAT$(change_total + change_count) + " file" + IIF$(change_total + change_count = 1,"","s") + "." '// keep user informed on progress
    mytick = RIGHT$(TIME$, 1)                                   '// remember this second
  END IF
  RETURN
END FUNCTION
